from blocks.types.multiprocess import ParallelEvent
from blocks.types.base import (
    Block,
    Event,
    Source,
    AnySource,
    Processor,
    AsyncSource,
    AnyProcessor,
    EventOrEvents,
    AsyncProcessor,
    TypeOfProcessor,
)
